import ast
import json

from requests import Response

ERROR_FLAG = "[ERROR]"
WARNING_FLAG = "[WARNING]"


def timing(func):
    def wrapper(*args, **kwargs):
        import time

        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        print(f"Finished {func.__name__.title()} in {int((end - start))} seconds")
        return result

    return wrapper


def remove_null_values_from_dict(source_dic: dict) -> dict:
    return {k: v for k, v in source_dic.items() if v is not None}


def read_response_data(resp: Response):
    return json.loads(resp.content).get("data", {})


def to_bool(value):
    try:
        return ast.literal_eval(value)
    except:
        return value


def is_jsonable(myjson):
    try:
        json.dumps(myjson)
    except ValueError:
        return False
    return True



import copy
import logging
import uuid
from pathlib import Path
from typing import Dict, List, Optional

import fsspec
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
import ray.data

import runhouse as rh
from runhouse.rh_config import rns_client
from runhouse.rns.folders.folder import folder
from .. import Resource, SkyCluster
from ..top_level_rns_fns import save

logger = logging.getLogger(__name__)


class Table(Resource):
    RESOURCE_TYPE = "table"
    DEFAULT_FOLDER_PATH = "/runhouse/tables"
    DEFAULT_CACHE_FOLDER = ".cache/runhouse/tables"
    DEFAULT_STREAM_FORMAT = "pyarrow"
    DEFAULT_BATCH_SIZE = 256
    DEFAULT_PREFETCH_BLOCKS = 1

    def __init__(
        self,
        url: str,
        name: Optional[str] = None,
        file_name: Optional[str] = None,
        fs: Optional[str] = None,
        data_config: Optional[dict] = None,
        dryrun: bool = True,
        partition_cols: Optional[List] = None,
        stream_format: Optional[str] = None,
        metadata: Optional[Dict] = None,
        **kwargs,
    ):
        super().__init__(name=name, dryrun=dryrun)
        self._filename = str(Path(url).name) if url else self.name
        # Use factory method so correct subclass for fs is returned
        self._folder = folder(url=url, fs=fs, data_config=data_config, dryrun=dryrun)
        self._cached_data = None
        self.partition_cols = partition_cols
        self.file_name = file_name
        self.stream_format = stream_format or self.DEFAULT_STREAM_FORMAT
        self.metadata = metadata or {}

    @staticmethod
    def from_config(config: dict, dryrun=True):
        if isinstance(config["fs"], dict):
            config["fs"] = SkyCluster.from_config(config["fs"], dryrun=dryrun)
        return Table(**config, dryrun=dryrun)

    @property
    def config_for_rns(self):
        config = super().config_for_rns
        if isinstance(self._folder, Resource):
            config["fs"] = self._resource_string_for_subconfig(self.fs)
            config["data_config"] = self._folder._data_config
        else:
            config["fs"] = self.fs
        self.save_attrs_to_config(config, ["url", "partition_cols", "metadata"])
        config.update(config)

        return config

    @property
    def data(self) -> ray.data.Dataset:
        """Get the table data. If data is not already cached return a ray dataset. With the dataset object we can
        stream or convert to other types, for example:
            data.iter_batches()
            data.to_pandas()
            data.to_dask()
        """
        if self._cached_data is not None:
            return self._cached_data
        return self.read_ray_dataset()

    @data.setter
    def data(self, new_data):
        """Update the data blob to new data"""
        self._cached_data = new_data
        # TODO should we save here?
        # self.save(overwrite=True)

    @property
    def fs(self):
        return self._folder.fs

    @fs.setter
    def fs(self, new_fs):
        self._folder.fs = new_fs

    @property
    def url(self):
        if self.file_name:
            return f"{self._folder.url}/{self.file_name}"
        return self._folder.url

    @url.setter
    def url(self, new_url):
        self._folder.url = new_url

    def set_metadata(self, key, val):
        self.metadata[key] = val

    def get_metadata(self, key):
        return self.metadata.get(key)

    @property
    def fsspec_url(self):
        if self.file_name:
            return f"{self._folder.fsspec_url}/{self.file_name}"
        return self._folder.fsspec_url

    @property
    def data_config(self):
        return self._folder.data_config

    @data_config.setter
    def data_config(self, new_data_config):
        self._folder.data_config = new_data_config

    def to(self, fs, url=None, data_config=None):
        new_table = copy.copy(self)
        new_table._folder = self._folder.to(fs=fs, url=url, data_config=data_config)
        return new_table

    def save(
        self,
        name: Optional[str] = None,
        snapshot: bool = False,
        overwrite: bool = True,
        **snapshot_kwargs,
    ):
        if self._cached_data is not None:
            data_to_write = self.data
            if isinstance(data_to_write, pa.Table):
                data_to_write = self.ray_dataset_from_arrow(data_to_write)

            if not isinstance(data_to_write, ray.data.Dataset):
                raise TypeError(f"Invalid Table format {type(data_to_write)}")

            self.write_ray_dataset(data_to_write)
            logger.info(f"Saved {str(self)} to: {self.fsspec_url}")

        save(self, snapshot=snapshot, overwrite=overwrite, **snapshot_kwargs)

        return self

    def fetch(self, columns: Optional[list] = None) -> pa.Table:
        # https://arrow.apache.org/docs/python/generated/pyarrow.parquet.read_table.html
        try:
            with fsspec.open(self.fsspec_url, mode="rb", **self.data_config) as t:
                self._cached_data = pq.read_table(t.full_name, columns=columns)
        except:
            # When trying to read as file like object could fail for a couple of reasons:
            # IsADirectoryError: The folder URL is actually a directory and the file has been automatically
            # generated for us inside the folder (ex: pyarrow table)

            # The file system is SFTP: since the SFTPFileSystem takes the host as a separate param, we cannot
            # pass in the data config as a single data_config kwarg

            # When specifying the filesystem don't pass in the fsspec url (which includes the file system prepended)
            self._cached_data = pq.read_table(
                self.url, columns=columns, filesystem=self._folder.fsspec_fs
            )
        return self._cached_data

    def __getstate__(self):
        """Override the pickle method to clear _cached_data before pickling"""
        state = self.__dict__.copy()
        state["_cached_data"] = None
        return state

    def __iter__(self):
        for block in self.stream(batch_size=self.DEFAULT_BATCH_SIZE):
            for sample in block:
                yield sample

    def __len__(self):
        if isinstance(self.data, pd.DataFrame):
            len_dataset = self.data.shape[0]

        elif isinstance(self.data, ray.data.Dataset):
            len_dataset = self.data.count()

        else:
            if not hasattr(self.data, "__len__") or not self.data:
                raise RuntimeError("Cannot get len for dataset.")
            else:
                len_dataset = len(self.data)

        return len_dataset

    def __str__(self):
        return self.__class__.__name__

    def stream(
        self,
        batch_size: int,
        drop_last: bool = False,
        shuffle_seed: Optional[int] = None,
        shuffle_buffer_size: Optional[int] = None,
        prefetch_blocks: Optional[int] = None,
    ):
        """Return a local batched iterator over the ray dataset.

        Args:
            batch_size: The number of rows in each batch. The final batch may include fewer than ``batch_size`` rows if
                ``drop_last`` is ``False``.
            drop_last: Whether to drop the last batch if it's incomplete.
            shuffle_seed: The seed to use for the local random shuffle.
            shuffle_buffer_size: minimum number of rows that must be in the local in-memory shuffle
                buffer in order to yield a batch. Must be greater than or equal to ``batch_size``.
            prefetch_blocks: The number of blocks to prefetch ahead of the current block during the scan.
        """
        ray_data = self.data

        if self.stream_format == "torch":
            # https://docs.ray.io/en/master/data/api/doc/ray.data.Dataset.iter_torch_batches.html#ray.data.Dataset.iter_torch_batches
            return ray_data.iter_torch_batches(
                batch_size=batch_size,
                prefetch_blocks=prefetch_blocks or self.DEFAULT_PREFETCH_BLOCKS,
                drop_last=drop_last,
                local_shuffle_buffer_size=shuffle_buffer_size,
                local_shuffle_seed=shuffle_seed,
            )

        elif self.stream_format == "tf":
            # https://docs.ray.io/en/master/data/api/doc/ray.data.Dataset.iter_tf_batches.html
            return ray_data.iter_tf_batches(
                batch_size=batch_size,
                prefetch_blocks=prefetch_blocks or self.DEFAULT_PREFETCH_BLOCKS,
                drop_last=drop_last,
                local_shuffle_buffer_size=shuffle_buffer_size,
                local_shuffle_seed=shuffle_seed,
            )
        else:
            # https://docs.ray.io/en/latest/data/api/dataset.html#ray.data.Dataset.iter_batches
            return ray_data.iter_batches(
                batch_size=batch_size,
                prefetch_blocks=prefetch_blocks or self.DEFAULT_PREFETCH_BLOCKS,
                batch_format=self.stream_format,
                drop_last=drop_last,
                local_shuffle_buffer_size=shuffle_buffer_size,
                local_shuffle_seed=shuffle_seed,
            )

    def read_ray_dataset(self, columns: Optional[List[str]] = None):
        """Read parquet data as a ray dataset object"""
        # https://docs.ray.io/en/latest/data/api/input_output.html#parquet
        dataset = ray.data.read_parquet(
            self.fsspec_url, columns=columns, filesystem=self._folder.fsspec_fs
        )
        return dataset

    def write_ray_dataset(self, data_to_write: ray.data.Dataset):
        """Write a ray dataset to a fsspec filesystem"""
        if self.partition_cols:
            # TODO [JL]: https://arrow.apache.org/docs/python/generated/pyarrow.parquet.write_to_dataset.html
            logger.warning("Partitioning by column not currently supported.")
            pass

        # https://docs.ray.io/en/master/data/api/doc/ray.data.Dataset.write_parquet.html
        data_to_write.write_parquet(self.fsspec_url, filesystem=self._folder.fsspec_fs)

    @staticmethod
    def ray_dataset_from_arrow(data):
        """Convert an Arrow Table to a ray Dataset"""
        return ray.data.from_arrow(data)

    def delete_in_fs(self, recursive: bool = True):
        """Remove contents of all subdirectories (ex: partitioned data folders)"""
        # If file(s) are directories, recursively delete contents and then also remove the directory
        # https://filesystem-spec.readthedocs.io/en/latest/api.html#fsspec.spec.AbstractFileSystem.rm
        self._folder.fsspec_fs.rm(self.url, recursive=recursive)

    def exists_in_fs(self):
        return self._folder.exists_in_fs() and len(self._folder.ls(self.fsspec_url)) > 1

    def from_cluster(self, cluster):
        """Create a remote folder from a url on a cluster. This will create a virtual link into the
        cluster's filesystem. If you want to create a local copy or mount of the folder, use
        `Folder(url=<local_url>).sync_from_cluster(<cluster>, <url>)` or
        `Folder('url').from_cluster(<cluster>).mount(<local_url>)`."""
        if not cluster.address:
            raise ValueError("Cluster must be started before copying data from it.")
        new_table = copy.deepcopy(self)
        new_table._folder.fs = cluster
        return new_table


def _load_table_subclass(data, config: dict, dryrun: bool):
    """Load the relevant Table subclass based on the config or data type provided"""
    resource_subtype = config.get("resource_subtype", Table.__name__)

    try:
        import datasets

        if isinstance(data, datasets.Dataset) or resource_subtype == "HuggingFaceTable":
            from .huggingface_table import HuggingFaceTable

            return HuggingFaceTable.from_config(config)
    except ModuleNotFoundError:
        pass
    except Exception as e:
        raise e

    try:
        if isinstance(data, pd.DataFrame) or resource_subtype == "PandasTable":
            from .pandas_table import PandasTable

            return PandasTable.from_config(config)
    except ModuleNotFoundError:
        pass
    except Exception as e:
        raise e

    try:
        import dask.dataframe as dd

        if isinstance(data, dd.DataFrame) or resource_subtype == "DaskTable":
            from .dask_table import DaskTable

            return DaskTable.from_config(config)
    except ModuleNotFoundError:
        pass
    except Exception as e:
        raise e

    try:
        import ray

        if isinstance(data, ray.data.Dataset) or resource_subtype == "RayTable":
            from .ray_table import RayTable

            return RayTable.from_config(config)
    except ModuleNotFoundError:
        pass
    except Exception as e:
        raise e

    try:
        import cudf

        if isinstance(data, cudf.DataFrame) or resource_subtype == "CudfTable":
            # TODO [JL]
            # from .rapids_table import RapidsTable

            # return RapidsTable.from_config(config)
            raise NotImplementedError("Cudf not currently supported")
    except ModuleNotFoundError:
        pass
    except Exception as e:
        raise e

    if isinstance(data, pa.Table) or resource_subtype == "Table":
        new_table = Table.from_config(config, dryrun=dryrun)
        return new_table
    else:
        raise TypeError(
            f"Unsupported data type {type(data)} for Table construction. "
            f"For converting data to pyarrow see: "
            f"https://arrow.apache.org/docs/7.0/python/generated/pyarrow.Table.html"
        )


def table(
    data=None,
    name: Optional[str] = None,
    url: Optional[str] = None,
    fs: Optional[str] = None,
    data_config: Optional[dict] = None,
    partition_cols: Optional[list] = None,
    mkdir: bool = False,
    dryrun: bool = False,
    stream_format: Optional[str] = None,
    metadata: Optional[dict] = None,
):
    """Returns a Table object, which can be used to interact with the table at the given url.
    If the table does not exist, it will be saved if `dryrun` is False.
    """
    config = rns_client.load_config(name)

    config["fs"] = fs or config.get("fs") or rns_client.DEFAULT_FS
    if isinstance(config["fs"], str) and rns_client.exists(
        config["fs"], resource_type="cluster"
    ):
        config["fs"] = rns_client.load_config(config["fs"])

    name = name or config.get("rns_address") or config.get("name")
    name = name.lstrip("/") if name is not None else name

    data_url = url or config.get("url")
    file_name = None
    if data_url:
        # Extract the file name from the url if provided
        full_path = Path(data_url)
        file_suffix = full_path.suffix
        if file_suffix:
            data_url = str(full_path.parent)
            file_name = full_path.name

    if data_url is None:
        # TODO [JL] move some of the default params in this factory method to the defaults module for configurability
        if config["fs"] == rns_client.DEFAULT_FS:
            # create random url to store in .cache folder of local filesystem
            data_url = str(
                Path(
                    f"~/{Table.DEFAULT_CACHE_FOLDER}/{name or uuid.uuid4().hex}"
                ).expanduser()
            )
        else:
            # save to the default bucket
            data_url = f"{Table.DEFAULT_FOLDER_PATH}/{name}"

    config["name"] = name
    config["url"] = data_url
    config["file_name"] = file_name or config.get("file_name")
    config["data_config"] = data_config or config.get("data_config")
    config["partition_cols"] = partition_cols or config.get("partition_cols")
    config["stream_format"] = stream_format or config.get("stream_format")
    config["metadata"] = metadata or config.get("metadata")

    if mkdir:
        # create the remote folder for the table
        rh.folder(url=data_url, fs=fs, dryrun=True).mkdir()

    new_table = _load_table_subclass(data, config, dryrun)
    if data is not None:
        new_table.data = data

    return new_table



import unittest
from pathlib import Path

import runhouse as rh
from runhouse.rh_config import rns_client


def test_find_working_dir(tmp_path):
    starting_dir = Path(tmp_path, "subdir/subdir/subdir/subdir")
    d = rns_client.locate_working_dir(cwd=str(starting_dir))
    assert d in str(starting_dir)

    Path(tmp_path, "subdir/rh").mkdir(parents=True)
    d = rns_client.locate_working_dir(str(starting_dir))
    assert d == str(Path(tmp_path, "subdir"))

    Path(tmp_path, "subdir/rh").rmdir()

    Path(tmp_path, "subdir/subdir/.git").mkdir(exist_ok=True, parents=True)
    d = rns_client.locate_working_dir(str(starting_dir))
    assert d in str(Path(tmp_path, "subdir/subdir"))

    Path(tmp_path, "subdir/subdir/.git").rmdir()

    Path(tmp_path, "subdir/subdir/requirements.txt").write_text("....")
    d = rns_client.locate_working_dir(str(starting_dir))
    assert d in str(Path(tmp_path, "subdir/subdir"))


def test_set_folder(tmp_path):
    rh.set_folder("~/tests")
    rh.folder(name="bert_ft").save()

    # TODO [DG] does this assume that the user must have runhouse in their home directory?
    assert rh.current_folder() == "~/tests"
    assert (Path(rh.rh_config.rns_client.rh_directory) / "tests/bert_ft").exists()
    assert rh.exists("~/tests/bert_ft")


def test_rns_path(tmp_path):
    rh.set_folder("~")
    assert rh.folder("tests").rns_address == "~/tests"

    rh.set_folder("@")
    assert (
        rh.folder("bert_ft").rns_address
        == rh.configs.get("default_folder") + "/bert_ft"
    )


def test_ls():
    rh.set_folder("~")
    assert rh.resources() == rh.resources("~")
    rh.set_folder("^")
    assert rh.resources() == [
        "rh-32-cpu",
        "rh-gpu",
        "rh-cpu",
        "rh-4-gpu",
        "rh-8-cpu",
        "rh-v100",
        "rh-8-v100",
        "rh-8-gpu",
        "rh-4-v100",
    ]
    assert (
        rh.resources("bert_ft") == []
    )  # We're still inside builtins so we can't see bert_ft
    assert rh.folder("~/tests", dryrun=False).resources() == ["bert_ft"]
    rh.set_folder("~")
    assert rh.folder("tests", dryrun=False).resources() == ["bert_ft"]
    assert rh.resources("tests") == ["bert_ft"]


def test_from_name():
    f = rh.Folder.from_name("~/bert_ft", dryrun=True)
    assert f.contains("my_test_hw")
    c = rh.SkyCluster.from_name("^rh-cpu", dryrun=True)
    assert c.instance_type == "m5.large"


if __name__ == "__main__":
    unittest.main()



import unittest
from pathlib import Path

import runhouse as rh


def setup():
    pass


def test_from_string():
    p = rh.Package.from_string("reqs:~/runhouse")
    assert p.local_path == str(Path.home() / "runhouse")


def test_share():
    p = rh.Package.from_string("reqs:~/runhouse/runhouse")
    # Can share via email or username
    p.share(users=["jlewitt1", "donnyg"], snapshot=True, access_type="write")


def test_reload():
    rh.exists("@/runhouse")
    package = rh.package(name="/donnyg/runhouse")
    print(package.config_for_rns)
    assert "s3://" in package.fsspec_url


if __name__ == "__main__":
    setup()
    unittest.main()



import json
import os
import unittest
from multiprocessing import Pool

import requests

import runhouse as rh
from runhouse.rns.api_utils.resource_access import ResourceAccess


def setup():
    rh.set_folder("~/tests", create=True)


def torch_summer(a, b):
    # import inside so tests that don't use torch don't fail because torch isn't in their reqs
    import torch

    return int(torch.Tensor([a, b]).sum())


def summer(a, b):
    return a + b


def test_create_send_from_name_local():
    local_sum = rh.send(
        fn=summer, name="local_send", hardware="^rh-cpu", reqs=["local:./"]
    ).save()
    del local_sum

    remote_sum = rh.send(name="local_send")
    res = remote_sum(1, 5)
    assert res == 6

    remote_sum.delete_configs()
    assert rh.exists("local_send") is False


def test_create_send_from_rns():
    remote_sum = rh.send(
        fn=summer, name="@/remote_send", hardware="^rh-cpu", reqs=[], dryrun=True
    ).save()
    del remote_sum

    remote_sum = rh.send(name="@/remote_send")
    res = remote_sum(1, 5)
    assert res == 6

    remote_sum.delete_configs()
    assert rh.exists("@/remote_send") is False


@unittest.skip("Not yet implemented.")
def test_running_send_as_proxy():
    remote_sum = rh.send(
        fn=summer, name="@/remote_send", hardware="^rh-cpu", reqs=[], dryrun=False
    ).save()
    del remote_sum

    remote_sum = rh.send(name="@/remote_send")
    remote_sum.access = ResourceAccess.proxy
    res = remote_sum(1, 5)
    assert res == 6

    remote_sum.delete_configs()
    assert rh.exists("@remote_send") is False


def test_get_send_history():
    remote_sum = rh.send(
        fn=summer, name="@/remote_send", hardware="^rh-cpu", reqs=[], dryrun=True
    ).save()
    remote_sum = rh.send(
        fn=summer, name="@/remote_send", hardware="^rh-cpu", reqs=["torch"], dryrun=True
    ).save()
    remote_sum = rh.send(
        fn=summer, name="@/remote_send", hardware="^rh-cpu", reqs=[], dryrun=True
    ).save()
    name = "@/remote_send"
    remote_sum = rh.send(name=name)
    history = remote_sum.history(name=name)
    assert len(history) >= 3
    assert "torch" in history[1]["data"]["reqs"]
    remote_sum.delete_configs()
    # TODO assert raises
    # history = remote_sum.history(name=name)
    # assert len(history) == 0


def multiproc_torch_sum(inputs):
    print(f"CPUs: {os.cpu_count()}")
    with Pool() as P:
        return P.starmap(torch_summer, inputs)


def test_remote_send_with_multiprocessing():
    re_fn = rh.send(
        multiproc_torch_sum,
        name="test_send",
        hardware="^rh-cpu",
        reqs=["./", "torch==1.12.1"],
    )
    summands = list(zip(range(5), range(4, 9)))
    res = re_fn(summands)
    assert res == [4, 6, 8, 10, 12]


def getpid(a=0):
    return os.getpid() + a


def test_maps():
    pid_fn = rh.send(getpid, hardware="^rh-cpu")
    num_pids = [1] * 50
    pids = pid_fn.map(num_pids)
    assert len(set(pids)) > 1

    pid_ref = pid_fn.remote()

    pids = pid_fn.repeat(num_repeats=50)
    assert len(set(pids)) > 1

    pids = [pid_fn.enqueue() for _ in range(10)]
    assert len(pids) == 10

    pid_res = pid_fn.get(pid_ref)
    assert pid_res > 0

    # Test passing an objectref into a normal call
    pid_res_from_ref = pid_fn(pid_ref)
    assert pid_res_from_ref > pid_res

    re_fn = rh.send(summer, hardware="^rh-cpu")
    summands = list(zip(range(5), range(4, 9)))
    res = re_fn.starmap(summands)
    assert res == [4, 6, 8, 10, 12]


def test_send_git_fn():
    remote_parse = rh.send(
        fn="https://github.com/huggingface/diffusers/blob/"
        "main/examples/dreambooth/train_dreambooth.py:parse_args",
        hardware="^rh-cpu",
        reqs=[
            "torch==1.12.1 --verbose",
            "torchvision==0.13.1",
            "transformers",
            "datasets",
            "evaluate",
            "accelerate",
            "pip:./diffusers --verbose",
        ],
    )
    args = remote_parse(
        input_args=[
            "--pretrained_model_name_or_path",
            "stabilityai/stable-diffusion-2-base",
            "--instance_data_dir",
            "remote_image_dir",
            "--instance_prompt",
            "a photo of sks person",
        ]
    )
    assert args.pretrained_model_name_or_path == "stabilityai/stable-diffusion-2-base"


@unittest.skip("Not working properly.")
def test_send_external_fn():
    """Test sending a module from reqs, not from working_dir"""
    import torch

    re_fn = rh.send(torch.sum, hardware="^rh-cpu", reqs=["torch"])
    res = re_fn(torch.arange(5))
    assert int(res) == 10


@unittest.skip("Runs indefinitely.")
def test_notebook():
    nb_sum = lambda x: multiproc_torch_sum(x)
    re_fn = rh.send(
        nb_sum, hardware="^rh-cpu", reqs=["./", "torch==1.12.1"], dryrun=True
    )
    re_fn.notebook()
    summands = list(zip(range(5), range(4, 9)))
    res = re_fn(summands)
    assert res == [4, 6, 8, 10, 12]
    re_fn.delete_configs()


@unittest.skip("Runs indefinitely.")
def test_ssh():
    # TODO do this properly
    my_send = rh.send(name="local_send")
    my_send.ssh()
    assert True


def test_providing_access_to_send():
    my_send = rh.send(
        fn=summer, name="@/remote_send", hardware="^rh-cpu", dryrun=True
    ).save()
    added_users, new_users = my_send.share(
        users=["dongreenberg", "jlewit1"], access_type=ResourceAccess.read
    )
    assert added_users or new_users
    my_send.delete_configs()


def delete_send_from_rns(s):
    server_url = s.rns_client.api_server_url
    resource_request_uri = s.rns_client.resource_uri(s.name)
    resp = requests.delete(
        f"{server_url}/resource/{resource_request_uri}",
        headers=s.rns_client.request_headers,
    )
    if resp.status_code != 200:
        raise Exception(
            f"Failed to delete_configs send data from url: {json.loads(resp.content)}"
        )

    try:
        # Terminate the cluster
        s.cluster.teardown_and_delete()
    except Exception as e:
        raise Exception(f"Failed to teardown the cluster: {e}")


@unittest.skip("Not yet implemented.")
def test_http_url():
    # TODO [DG] shouldn't have to specify fn here as a callable / at all?
    s = rh.send(fn=summer, name="test_send", hardware="^rh-cpu")

    # Generate and call the URL
    http_url = s.http_url()
    assert http_url

    res = s(a=1, b=2)

    # delete_configs the Send data from the RNS
    # delete_send_from_rns(s)

    assert res == 3


@unittest.skip("Not yet implemented.")
def test_http_url_with_curl():
    # NOTE: Assumes the Send has already been created and deployed to running cluster
    s = rh.send(name="test_send")
    curl_cmd = s.http_url(a=1, b=2, curl_command=True)
    print(curl_cmd)

    # delete_configs the send data from the RNS
    delete_send_from_rns(s)

    assert True


def test_byo_cluster_send():
    # Spin up a new basic m5.xlarge EC2 instance
    c = rh.cluster(
        instance_type="m5.xlarge",
        provider="aws",
        region="us-east-1",
        image_id="ami-0a313d6098716f372",
        name="test-byo-cluster",
    ).up_if_not()
    ip = c.address
    creds = c.ssh_creds()
    del c
    byo_cluster = rh.cluster(name="different-cluster", ips=[ip], ssh_creds=creds).save()
    re_fn = rh.send(
        multiproc_torch_sum, hardware=byo_cluster, reqs=["./", "torch==1.12.1"]
    )
    summands = list(zip(range(5), range(4, 9)))
    res = re_fn(summands)
    assert res == [4, 6, 8, 10, 12]


def test_byo_cluster_maps():
    pid_fn = rh.send(getpid, hardware="different-cluster")
    num_pids = [1] * 50
    pids = pid_fn.map(num_pids)
    assert len(set(pids)) > 1

    pid_ref = pid_fn.remote()

    pids = pid_fn.repeat(num_repeats=50)
    assert len(set(pids)) > 1

    pids = [pid_fn.enqueue() for _ in range(10)]
    assert len(pids) == 10

    pid_res = pid_fn.get(pid_ref)
    assert pid_res > 0

    # Test passing an objectref into a normal call
    pid_res_from_ref = pid_fn(pid_ref)
    assert pid_res_from_ref > pid_res

    re_fn = rh.send(summer, hardware="different-cluster")
    summands = list(zip(range(5), range(4, 9)))
    res = re_fn.starmap(summands)
    assert res == [4, 6, 8, 10, 12]


if __name__ == "__main__":
    setup()
    unittest.main()



import copy
import logging
import re
import subprocess
import sys
from pathlib import Path
from typing import Dict, Optional, Union

from runhouse import rh_config
from runhouse.rns.folders.folder import Folder
from runhouse.rns.resource import Resource

INSTALL_METHODS = {"local", "reqs", "pip", "conda"}


class Package(Resource):
    RESOURCE_TYPE = "package"

    def __init__(
        self,
        name: str = None,
        install_method: str = None,
        install_target: Union[str, Folder] = None,
        install_args: str = None,
        dryrun: bool = False,
        **kwargs,  # We have this here to ignore extra arguments when calling from from_config
    ):
        super().__init__(
            name=name,
            dryrun=dryrun,
        )
        self.install_method = install_method
        self.install_target = install_target
        self.install_args = install_args

    @property
    def config_for_rns(self):
        # If the package is just a simple Package.from_string string, no
        # need to store it in rns, just give back the string.
        # if self.install_method in ['pip', 'conda', 'git']:
        #     return f'{self.install_method}:{self.name}'
        config = super().config_for_rns
        config["install_method"] = self.install_method
        config["install_target"] = self._resource_string_for_subconfig(
            self.install_target
        )
        config["install_args"] = self.install_args
        return config

    def __str__(self):
        if self.name:
            return f"Package: {self.name}"
        if isinstance(self.install_target, Folder):
            # if self.install_target.name:
            #     return f'Package: {self.install_target.name}'
            return f"Package: {self.install_target.url}"
        return f"Package: {self.install_target}"

    def install(self):
        logging.info(
            f"Installing package {str(self)} with method {self.install_method}."
        )
        install_cmd = ""
        install_args = f" {self.install_args}" if self.install_args else ""
        if isinstance(self.install_target, Folder):
            local_path = self.install_target.local_path
            if not local_path:
                local_path = "~/" + self.name
            elif not self.install_target.is_local():
                # TODO [DG] replace this with empty mount() call to be put in tmp folder by Folder
                local_path = self.install_target.mount(
                    url=f"~/{Path(self.install_target.url).stem}"
                )

            if self.install_method == "pip":
                # TODO [DG] Revisit: Would be nice if we could use -e by default, but importlib on the grpc server
                #  isn't finding the package right after its installed.
                # if (Path(local_path) / 'setup.py').exists():
                #     install_cmd = f'-e {local_path}' + install_args
                # else:
                install_cmd = f"{local_path}" + install_args
            elif self.install_method == "conda":
                install_cmd = f"{local_path}" + install_args
            elif self.install_method == "reqs":
                if (Path(local_path) / "requirements.txt").exists():
                    logging.info(
                        f"Attempting to install requirements from {local_path}/requirements.txt"
                    )
                    self.pip_install(
                        f"-r {Path(local_path)}/requirements.txt" + install_args
                    )
                else:
                    logging.info(f"{local_path}/requirements.txt not found, skipping")
        else:
            install_cmd = self.install_target + install_args

        if self.install_method == "pip":
            self.pip_install(install_cmd)
        elif self.install_method == "conda":
            self.conda_install(install_cmd)
        elif self.install_method in ["local", "reqs"]:
            if isinstance(self.install_target, Folder):
                sys.path.append(local_path)
            elif Path(self.install_target).resolve().expanduser().exists():
                sys.path.append(str(Path(self.install_target).resolve().expanduser()))
            else:
                raise ValueError(
                    f"install_target must be a Folder or a path to a directory for "
                    f"install_method {self.install_method}"
                )
        # elif self.install_method == 'unpickle':
        #     # unpickle the functions and make them importable
        #     with self.get('functions.pickle') as f:
        #         sys.modules[self.name] = pickle.load(f)
        else:
            raise ValueError(
                f"Unknown install_method {self.install_method}. Try using cluster.run() or "
                f"send.run_setup() to install instead."
            )

    @staticmethod
    def pip_install(install_cmd: str):
        logging.info(f"Running: pip install {install_cmd}")
        subprocess.check_call(
            [sys.executable, "-m", "pip", "install"] + install_cmd.split(" ")
        )

    @staticmethod
    def conda_install(install_cmd: str):
        logging.info(f"Running: conda install {install_cmd}")
        # check if conda is installed, and if not, install it
        try:
            subprocess.check_call(["conda", "--version"])
            subprocess.run(["conda", "install", "-y"] + install_cmd.split(" "))
        except FileNotFoundError:
            logging.info("Conda not found, installing...")
            subprocess.check_call(
                "wget https://repo.continuum.io/miniconda/Miniconda3-latest-Linux-x86_64.sh "
                "-O ~/miniconda.sh".split(" ")
            )
            subprocess.check_call(["bash", "~/miniconda.sh", "-b", "-p", "~/miniconda"])
            subprocess.check_call("source $HOME/miniconda3/bin/activate".split(" "))
            status = subprocess.check_call(
                ["conda", "install", "-y"] + install_cmd.split(" ")
            )
            if not status == 0:
                raise RuntimeError(
                    "Conda install failed, check that the package exists and is "
                    "available for your platform."
                )

    def to_cluster(self, dest_cluster, url=None, mount=False, return_dest_folder=False):
        if isinstance(self.install_target, Folder):
            new_folder = self.install_target.to_cluster(
                dest_cluster,
                url=url,
                mount=mount,
                return_dest_folder=return_dest_folder,
            )
            new_package = copy.copy(self)
            new_package.install_target = new_folder
            return new_package

    @staticmethod
    def from_config(config: dict, dryrun=False):
        if config.resource_subtype == "GitPackage":
            from git_package import GitPackage

            return GitPackage.from_config(config, dryrun=dryrun)
        return Package(**config, dryrun=dryrun)

    @staticmethod
    def from_string(specifier: str, dryrun=False):
        # Use regex to check if specifier matches '<method>:https://github.com/<path>' or 'https://github.com/<path>'
        match = re.search(r"(?:<(.*)>:)?https://github.com/(.*)", specifier)
        if match:
            install_method = match.group(1)
            url = match.group(2)
            from runhouse.rns.packages.git_package import git_package

            return git_package(
                git_url=url, install_method=install_method, dryrun=dryrun
            )

        target_and_args = specifier
        if specifier.split(":")[0] in INSTALL_METHODS:
            target_and_args = specifier.split(":", 1)[1]
        rel_target, args = (
            target_and_args.split(" ", 1)
            if " " in target_and_args
            else (target_and_args, "")
        )
        # We need to do this because relative paths are relative to the current working directory!
        abs_target = (
            Path(rel_target).expanduser()
            if Path(rel_target).expanduser().is_absolute()
            else Path(rh_config.rns_client.locate_working_dir()) / rel_target
        )
        if abs_target.exists():
            target = Folder(url=rel_target)
        else:
            target = rel_target

        if specifier.startswith("local:"):
            return Package(install_target=target, install_method="local", dryrun=dryrun)
        elif specifier.startswith("reqs:"):
            return Package(
                install_target=target,
                install_args=args,
                install_method="reqs",
                dryrun=dryrun,
            )
        elif specifier.startswith("git+"):
            return Package(
                install_target=specifier[4:],
                install_args=args,
                install_method="pip",
                dryrun=dryrun,
            )
        elif specifier.startswith("pip:"):
            return Package(
                install_target=target,
                install_args=args,
                install_method="pip",
                dryrun=dryrun,
            )
        elif specifier.startswith("conda:"):
            return Package(
                install_target=target,
                install_args=args,
                install_method="conda",
                dryrun=dryrun,
            )
        elif specifier.startswith("rh:"):
            # Calling the factory function below
            return package(name=specifier[3:], dryrun=dryrun)
        else:
            if Path(specifier).resolve().exists():
                return Package(
                    install_target=target,
                    install_args=args,
                    install_method="reqs",
                    dryrun=dryrun,
                )
            else:
                return Package(
                    install_target=target,
                    install_args=args,
                    install_method="pip",
                    dryrun=dryrun,
                )


def package(
    name=None,
    install_method=None,
    install_str=None,
    url=None,
    fs: Optional[str] = Folder.DEFAULT_FS,
    dryrun=False,
    local_mount: bool = False,
    data_config: Optional[Dict] = None,
):
    config = rh_config.rns_client.load_config(name)
    config["name"] = name or config.get("rns_address", None) or config.get("name")

    config["install_method"] = install_method or config.get("install_method")
    if url is not None:
        config["install_target"] = Folder(
            url=url, fs=fs, local_mount=local_mount, data_config=data_config
        )
        config["install_args"] = install_str
    elif install_str is not None:
        config["install_target"], config["install_args"] = install_str.split(" ", 1)
    elif "install_target" in config and isinstance(config["install_target"], dict):
        config["install_target"] = Folder.from_config(config["install_target"])

    new_package = Package.from_config(config, dryrun=dryrun)

    return new_package



import importlib
import logging
from datetime import datetime
from functools import wraps
from pathlib import Path

import ray

from runhouse import rh_config


logger = logging.getLogger(__name__)


def call_fn_by_type(fn, fn_type, fn_name, module_path=None, args=None, kwargs=None):
    run_key = f"{fn_name}_{datetime.now().strftime('%Y%m%d_%H%M%S')}"

    # TODO other possible fn_types: 'batch', 'streaming'
    if fn_type == "call":
        args = [
            rh_config.obj_store.get(arg, default=arg) if isinstance(arg, str) else arg
            for arg in args
        ]
        kwargs = {
            k: rh_config.obj_store.get(v, default=v) if isinstance(v, str) else v
            for k, v in kwargs.items()
        }
        res = fn(*args, **kwargs)
    elif fn_type == "get":
        obj_ref = args[0]
        res = rh_config.obj_store.get(obj_ref)
    else:
        args = rh_config.obj_store.get_obj_refs_list(args)
        kwargs = rh_config.obj_store.get_obj_refs_dict(kwargs)

        logging_wrapped_fn = enable_logging_fn_wrapper(fn, run_key)

        has_gpus = ray.cluster_resources().get("GPU", 0) > 0

        # We need to add the module_path to the PYTHONPATH because ray runs remotes in a new process
        # We need to set max_calls to make sure ray doesn't cache the remote function and ignore changes to the module
        # See: https://docs.ray.io/en/releases-2.2.0/ray-core/package-ref.html#ray-remote
        # We need non-zero cpus and gpus for Ray to allow access to the compute.
        # We should see if there's a more elegant way to specify this.
        ray_fn = ray.remote(
            num_cpus=0.0001,
            num_gpus=0.0001 if has_gpus else None,
            max_calls=len(args) if fn_type in ["map", "starmap"] else 1,
            runtime_env={"env_vars": {"PYTHONPATH": module_path or ""}},
        )(logging_wrapped_fn)
        if fn_type == "map":
            obj_ref = [ray_fn.remote(arg, **kwargs) for arg in args]
        elif fn_type == "starmap":
            obj_ref = [ray_fn.remote(*arg, **kwargs) for arg in args]
        elif fn_type == "queue" or fn_type == "remote":
            obj_ref = ray_fn.remote(*args, **kwargs)
        elif fn_type == "repeat":
            [num_repeats, args] = args
            obj_ref = [ray_fn.remote(*args, **kwargs) for _ in range(num_repeats)]
        else:
            raise ValueError(f"fn_type {fn_type} not recognized")

        if fn_type == "remote":
            rh_config.obj_store.put_obj_ref(key=run_key, obj_ref=obj_ref)
            res = run_key
        else:
            res = ray.get(obj_ref)
    return res


def get_fn_by_name(module_name, fn_name):
    if module_name in rh_config.obj_store.imported_modules:
        importlib.invalidate_caches()
        rh_config.obj_store.imported_modules[module_name] = importlib.reload(
            rh_config.obj_store.imported_modules[module_name]
        )
        logger.info(f"Reloaded module {module_name}")
    else:
        rh_config.obj_store.imported_modules[module_name] = importlib.import_module(
            module_name
        )
        logger.info(f"Importing module {module_name}")
    fn = getattr(rh_config.obj_store.imported_modules[module_name], fn_name)
    return fn


RAY_LOGFILE_PATH = "/tmp/ray/session_latest/logs"


def enable_logging_fn_wrapper(fn, run_key):
    @wraps(fn)
    def wrapped_fn(*inner_args, **inner_kwargs):
        """Sets the logfiles for a function to be the logfiles for the current ray worker.
        This is used to stream logs from the worker to the client."""
        worker_id = ray._private.worker.global_worker.worker_id.hex()
        logdir = Path.home() / ".rh/logs" / run_key
        logdir.mkdir(exist_ok=True, parents=True)
        ray_logs_path = Path(RAY_LOGFILE_PATH)
        stdout_files = list(ray_logs_path.glob(f"worker-{worker_id}-*"))
        # Create simlinks to the ray log files in the run directory
        logger.info(f"Writing logs on cluster to {logdir}")
        for f in stdout_files:
            symlink = logdir / f.name
            if not symlink.exists():
                symlink.symlink_to(f)
        return fn(*inner_args, **inner_kwargs)

    return wrapped_fn



import logging
import sys
import traceback
from concurrent import futures
from pathlib import Path

import grpc

import ray
import ray.cloudpickle as pickle
from sky.skylet.autostop_lib import set_last_active_time_to_now

import runhouse.grpc_handler.unary_pb2 as pb2

import runhouse.grpc_handler.unary_pb2_grpc as pb2_grpc
from runhouse.grpc_handler.unary_client import OutputType
from runhouse.rh_config import obj_store
from runhouse.rns.packages.package import Package
from runhouse.rns.run_module_utils import call_fn_by_type, get_fn_by_name
from runhouse.rns.top_level_rns_fns import (
    clear_pinned_memory,
    pinned_keys,
    remove_pinned_object,
)

logger = logging.getLogger(__name__)


class UnaryService(pb2_grpc.UnaryServicer):
    DEFAULT_PORT = 50052
    MAX_MESSAGE_LENGTH = 1 * 1024 * 1024 * 1024  # 1 GB
    LOGGING_WAIT_TIME = 1.0

    def __init__(self, *args, **kwargs):
        ray.init(address="auto")
        self._installed_packages = []
        self.register_activity()

    def register_activity(self):
        set_last_active_time_to_now()

    def InstallPackages(self, request, context):
        self.register_activity()
        # logger.info(f"Message received from client: {request.message}")
        packages = pickle.loads(request.message)
        logger.info(f"Message received from client to install packages: {packages}")
        for package in packages:
            if isinstance(package, str):
                pkg = Package.from_string(package)
            elif hasattr(package, "install"):
                pkg = package
            else:
                raise ValueError(f"package {package} not recognized")

            if (str(pkg)) in self._installed_packages:
                continue
            logger.info(f"Installing package: {str(pkg)}")
            pkg.install()
            self._installed_packages.append(str(pkg))

        self.register_activity()
        return pb2.MessageResponse(message=pickle.dumps(True), received=True)

    def GetObject(self, request, context):
        self.register_activity()
        key, stream_logs = pickle.loads(request.message)
        logger.info(f"Message received from client to get object: {key}")
        if not stream_logs:
            return pb2.MessageResponse(
                message=pickle.dumps(obj_store.get(key)),
                received=True,
                output_type=OutputType.RESULT,
            )

        logfiles = None
        open_files = None
        ret_obj = None
        returned = False
        while not returned:
            try:
                res = obj_store.get(key, timeout=self.LOGGING_WAIT_TIME)
                logger.info(f"Got object of type {type(res)} back from object store")
                ret_obj = [res, None, None]
                returned = True
                # Don't return yet, go through the loop once more to get any remaining log lines
            except ray.exceptions.GetTimeoutError:
                pass
            except ray.exceptions.TaskCancelledError as e:
                logger.info(f"Attempted to get task {key} that was cancelled.")
                returned = True
                ret_obj = [None, e, traceback.format_exc()]

            if not logfiles:
                logfiles = obj_store.get_logfiles(key)
                open_files = [open(i, "r") for i in logfiles]
                logger.info(f"Streaming logs for {key} from {logfiles}")

            # Grab all the lines written to all the log files since the last time we checked
            ret_lines = []
            for i, f in enumerate(open_files):
                file_lines = f.readlines()
                if file_lines:
                    # TODO [DG] handle .out vs .err, and multiple workers
                    # if len(logfiles) > 1:
                    #     ret_lines.append(f"Process {i}:")
                    ret_lines += file_lines
            if ret_lines:
                yield pb2.MessageResponse(
                    message=pickle.dumps(ret_lines),
                    received=True,
                    output_type=OutputType.STDOUT,
                )

        # We got the object back from the object store, so we're done (but we went through the loop once
        # more to get any remaining log lines)
        [f.close() for f in open_files]
        yield pb2.MessageResponse(
            message=pickle.dumps(ret_obj), received=True, output_type=OutputType.RESULT
        )

    def ClearPins(self, request, context):
        self.register_activity()
        pins_to_clear = pickle.loads(request.message)
        logger.info(
            f"Message received from client to clear pins: {pins_to_clear or 'all'}"
        )
        cleared = []
        if pins_to_clear:
            for pin in pins_to_clear:
                remove_pinned_object(pin)
                cleared.append(pin)
        else:
            cleared = list(pinned_keys())
            clear_pinned_memory()

        self.register_activity()
        return pb2.MessageResponse(message=pickle.dumps(cleared), received=True)

    def CancelRun(self, request, context):
        self.register_activity()
        run_keys, force = pickle.loads(request.message)
        if not hasattr(run_keys, "len"):
            run_keys = [run_keys]
        obj_refs = obj_store.get_obj_refs_list(run_keys)
        [
            ray.cancel(obj_ref, force=force, recursive=True)
            for obj_ref in obj_refs
            if isinstance(obj_ref, ray.ObjectRef)
        ]
        return pb2.MessageResponse(
            message=pickle.dumps("Cancelled"),
            received=True,
            output_type=OutputType.RESULT,
        )

    def RunModule(self, request, context):
        self.register_activity()
        # get the function result from the incoming request
        try:
            [relative_path, module_name, fn_name, fn_type, args, kwargs] = pickle.loads(
                request.message
            )

            module_path = None
            # If relative_path is None, the module is not in the working dir, and should be in the reqs
            if relative_path:
                module_path = str((Path.home() / relative_path).resolve())
                sys.path.append(module_path)
                logger.info(f"Appending {module_path} to sys.path")

            if module_name == "notebook":
                fn = fn_name  # Already unpickled above
            else:
                fn = get_fn_by_name(module_name, fn_name)

            res = call_fn_by_type(fn, fn_type, fn_name, module_path, args, kwargs)
            # [res, None, None] is a silly hack for packaging result alongside exception and traceback
            result = {"message": pickle.dumps([res, None, None]), "received": True}

            self.register_activity()
            return pb2.MessageResponse(**result)
        except Exception as e:
            logger.exception(e)
            message = [None, e, traceback.format_exc()]
            self.register_activity()
            return pb2.MessageResponse(message=pickle.dumps(message), received=False)


def serve():
    server = grpc.server(
        futures.ThreadPoolExecutor(max_workers=10),
        options=[
            ("grpc.max_send_message_length", UnaryService.MAX_MESSAGE_LENGTH),
            ("grpc.max_receive_message_length", UnaryService.MAX_MESSAGE_LENGTH),
        ],
    )
    pb2_grpc.add_UnaryServicer_to_server(UnaryService(), server)
    server.add_insecure_port("[::]:50052")
    server.start()
    logger.info("Server up and running")
    server.wait_for_termination()


if __name__ == "__main__":
    serve()
